// *****************************************************************************
//
// Title       : dp128_aes_datapath
// Creator     : Øivind Ekelund
// Date        : 25.03.09
//
// Description : Datapath of the AES module
//               Datapath is split into two main parts data and key.
//               The data part operates on 32 bits at a time, and performs the
//               4 AES operations, ShiftRows, SubBytes (sboxes), MixColumns and
//               AddRoundKey(xor)
//               The key path also utilizes the sboxes in the rounds where it is
//               not used by the data path. The round constant is changed each
//               round and used in the key expansion.
//
// *****************************************************************************

module dp128_aes_datapath
  (
   input wire  [127:0]    data_in,
   input wire  [127:0]    key_in,
   input wire             decrypt,
   input wire             sb_mux,
   input wire             dpout_mux,
   input wire             addrk_mux,
   input wire  [1:0]      shrow_mux,
   input wire  [7:0]      rcon,
   output wire [127:0]    data_out,
   output wire [127:0]    key_out
   );
   
   // *****************************************************************************
   // Signal declarations
   // *****************************************************************************
   
   wire [127:0]           data_invshiftrow;
   wire [127:0]           data_shiftrow;
   wire [127:0]           rows_shifted;
   wire [127:0]           mc_out;
   wire [127:0]           addrk_in;
   wire [127:0]           addrk_out;
   wire [31:0]            key_in_31_0_rot;
   wire [127:0]           sb_out;
   wire [31:0]            rcon_out;
   
   reg [127:0]            datapath32_out;
   reg [127:0]            sb_in;
   
   wire [31:0]            key_tosb;
   wire [31:0]            keysb_out;
   
   // *****************************************************************************
   // Logics for roundkey computation
   // *****************************************************************************
   
   dp128_aes_keypath 
     u_keypath(
               .key_in   (key_in),
               .from_sb  (rcon_out),
               .decrypt  (decrypt),
               .key_out  (key_out),
               .to_sb    (key_tosb)
               );
               
   //Rotate keywords
   assign key_in_31_0_rot = {key_tosb[23:16], key_tosb[15:8], key_tosb[7:0], key_tosb[31:24]};               

   //Sboxes used in roundkey computation
   dp128_aes_sbox 
      u0_ksbox(
               .sb_in   (key_in_31_0_rot[7:0]),
               .inverse (1'b0),
               .sb_out  (keysb_out[7:0])
               );
    dp128_aes_sbox 
      u1_ksbox(
               .sb_in   (key_in_31_0_rot[15:8]),
               .inverse (1'b0),
               .sb_out  (keysb_out[15:8])
               );
    dp128_aes_sbox 
      u2_ksbox(
               .sb_in   (key_in_31_0_rot[23:16]),
               .inverse (1'b0),
               .sb_out  (keysb_out[23:16])
               );
    dp128_aes_sbox 
      u3_ksbox(
               .sb_in   (key_in_31_0_rot[31:24]),
               .inverse (1'b0),
               .sb_out  (keysb_out[31:24])
               );
               
   //Add roundconstant
   assign rcon_out = keysb_out ^ {rcon, 24'h000000};         
   
   // *****************************************************************************
   // MixColumns and InvMixColumns
   // *****************************************************************************
   
   dp128_aes_mixcolumns 
     u0_mixcolumns(
                  .mc_in   (data_in[31:0]),
                  .roundkey(key_in[31:0]),
                  .inverse (decrypt),
                  .mc_out  (mc_out[31:0])
                  );
                  
   dp128_aes_mixcolumns 
     u1_mixcolumns(
                  .mc_in   (data_in[63:32]),
                  .roundkey(key_in[63:32]),
                  .inverse (decrypt),
                  .mc_out  (mc_out[63:32])
                  );
                  
   dp128_aes_mixcolumns 
     u2_mixcolumns(
                  .mc_in   (data_in[95:64]),
                  .roundkey(key_in[95:64]),
                  .inverse (decrypt),
                  .mc_out  (mc_out[95:64])
                  );
                  
   dp128_aes_mixcolumns 
     u3_mixcolumns(
                  .mc_in   (data_in[127:96]),
                  .roundkey(key_in[127:96]),
                  .inverse (decrypt),
                  .mc_out  (mc_out[127:96])
                  );

   // *****************************************************************************
   // Sboxes
   // ***************************************************************************** 
   
   dp128_aes_sbox 
     u0_sbox(
             .sb_in   (sb_in[7:0]),
             .inverse (decrypt),
             .sb_out  (sb_out[7:0])
             );
   dp128_aes_sbox 
     u1_sbox(
             .sb_in   (sb_in[15:8]),
             .inverse (decrypt),
             .sb_out  (sb_out[15:8])
             );
   dp128_aes_sbox 
     u2_sbox(
             .sb_in(sb_in[23:16]),
             .inverse(decrypt),
             .sb_out(sb_out[23:16])
             );
   dp128_aes_sbox 
     u3_sbox(
             .sb_in   (sb_in[31:24]),
             .inverse (decrypt),
             .sb_out  (sb_out[31:24])
             );
             
   dp128_aes_sbox 
     u4_sbox(
             .sb_in   (sb_in[39:32]),
             .inverse (decrypt),
             .sb_out  (sb_out[39:32])
             );
   dp128_aes_sbox 
     u5_sbox(
             .sb_in   (sb_in[47:40]),
             .inverse (decrypt),
             .sb_out  (sb_out[47:40])
             );
   dp128_aes_sbox 
     u6_sbox(
             .sb_in(sb_in[55:48]),
             .inverse(decrypt),
             .sb_out(sb_out[55:48])
             );
   dp128_aes_sbox 
     u7_sbox(
             .sb_in   (sb_in[63:56]),
             .inverse (decrypt),
             .sb_out  (sb_out[63:56])
             );
    
   dp128_aes_sbox 
     u8_sbox(
             .sb_in   (sb_in[71:64]),
             .inverse (decrypt),
             .sb_out  (sb_out[71:64])
             );
   dp128_aes_sbox 
     u9_sbox(
             .sb_in   (sb_in[79:72]),
             .inverse (decrypt),
             .sb_out  (sb_out[79:72])
             );
   dp128_aes_sbox 
     u10_sbox(
             .sb_in(sb_in[87:80]),
             .inverse(decrypt),
             .sb_out(sb_out[87:80])
             );
   dp128_aes_sbox 
     u11_sbox(
             .sb_in   (sb_in[95:88]),
             .inverse (decrypt),
             .sb_out  (sb_out[95:88])
             );
             
   dp128_aes_sbox 
     u12_sbox(
             .sb_in   (sb_in[103:96]),
             .inverse (decrypt),
             .sb_out  (sb_out[103:96])
             );
   dp128_aes_sbox 
     u13_sbox(
             .sb_in   (sb_in[111:104]),
             .inverse (decrypt),
             .sb_out  (sb_out[111:104])
             );
   dp128_aes_sbox 
     u14_sbox(
             .sb_in(sb_in[119:112]),
             .inverse(decrypt),
             .sb_out(sb_out[119:112])
             );
   dp128_aes_sbox 
     u15_sbox(
             .sb_in   (sb_in[127:120]),
             .inverse (decrypt),
             .sb_out  (sb_out[127:120])
             );

   // *****************************************************************************
   // Muxes and xor-gates in the datapath
   // ***************************************************************************** 
   

   
   //invShiftrows
   assign data_invshiftrow = 
                             {
                              datapath32_out[127:120], datapath32_out[23:16],   datapath32_out[47:40],   datapath32_out[71:64],
                              datapath32_out[95:88],   datapath32_out[119:112], datapath32_out[15:8],    datapath32_out[39:32],
                              datapath32_out[63:56],   datapath32_out[87:80],   datapath32_out[111:104], datapath32_out[7:0],
                              datapath32_out[31:24],   datapath32_out[55:48],   datapath32_out[79:72],   datapath32_out[103:96]
                              };
   
   //shiftrows
   assign data_shiftrow = 
                          {
                           datapath32_out[127:120], datapath32_out[87:80],   datapath32_out[47:40],   datapath32_out[7:0],
                           datapath32_out[95:88],   datapath32_out[55:48],   datapath32_out[15:8],    datapath32_out[103:96],
                           datapath32_out[63:56],   datapath32_out[23:16],   datapath32_out[111:104], datapath32_out[71:64],
                           datapath32_out[31:24],   datapath32_out[119:112], datapath32_out[79:72],   datapath32_out[39:32]
                           };
   

   //Mux, choosing input to the Sbox
   always @ * 
     begin
        case(sb_mux)
          1'b0: sb_in   = addrk_out;
          1'b1: sb_in   = mc_out;
        endcase
     end
   
   //Mux, choosing output from the 32-bit datapath
   always @ * 
     begin
        case(dpout_mux)
          1'b0: datapath32_out   = sb_out;
          1'b1: datapath32_out   = sb_in;
        endcase
     end
   
   
   //Mux, choosing input to Addroundkey
   assign addrk_in = addrk_mux ? data_in : mc_out;
   
   //Addroundkey
   assign addrk_out = addrk_in ^ key_in;
   
   //Mux, choosing between shiftrows and invshiftrows
   assign data_out = (shrow_mux == 2'b00) ? data_shiftrow : (shrow_mux == 2'b01) ? data_invshiftrow : datapath32_out;
   
   
endmodule
