// *****************************************************************************
//
// Title       : dp128_aes_sequencer
// Creator     : Øivind Ekelund
// Date        : 25.03.09
//
// Description : Sequencer, controlling the AES datapath and registers
//
// *****************************************************************************

module dp128_aes_sequencer
  (
   input wire          clk_core,
   input wire          clken_core,
   input wire          reset_core_n,
   input wire          start,
   input wire          decrypt,
   //input wire          aes256,
   input wire          stop,
   //input wire          keybufen,
   output reg          running,
   output reg          addrk_mux,
   output reg          sb_mux,
   output reg          dpout_mux,
   output reg [1:0]    shrow_mux,
   output reg [7:0]    rcon,
   output reg [3:0]    data_we,
   output reg [3:0]    key_we,
   output reg          done
   );
   
   // *****************************************************************************
   // States in the FSM
   // *****************************************************************************
   
   parameter [1:0]

                IDLE             = 2'b00, 
                 
                ENC128_INIT_RND  = 2'b01,
                ENC128_RND       = 2'b10,
                ENC128_FINAL     = 2'b11;


   // *****************************************************************************
   // Register- and signal declarations
   // *****************************************************************************
   
   reg [1:0]           state;
   reg [1:0]           next_state;
   reg [3:0]           round_cnt;
   reg                 cnt_up;
   reg                 reset_cnt; 
   
   wire aes256,keybufen;
   assign aes256 = 0;
   assign keybufen = 0;
   
   // *****************************************************************************
   // Combinational part of the FSM
   // *****************************************************************************
   always @ * 
     begin
        //default values:
        //Controlsignals for the datapath
        running      = 1'b1;
        addrk_mux    = 1'b0;
        sb_mux       = 1'b0;
        dpout_mux    = 1'b0;
        shrow_mux    = 2'b00;   
        data_we      = 4'h0;
        key_we       = 4'h0;
        done         = 1'b0;
        
        //Internal controlsignals
        next_state = state;
        cnt_up     = 1'b0;
        reset_cnt  = 1'b0;
        
        case (state)
          IDLE: 
            begin
               running   = 1'b0;
               
               if (start)
                 begin
                    reset_cnt = 1'b1; 
                    
                    next_state = ENC128_INIT_RND;
                 end
            end
          
          ENC128_INIT_RND: 
            begin
               //Initial round of AES128 encryption
               //SubBytes and Addroundkey is perfomed in this state
               
               if (decrypt) shrow_mux = 2'b01; // Inverse shiftrows
                
               sb_mux    = 1'b0;             // Do Addroundkey
               addrk_mux = 1'b1;             // No MixColumns
               dpout_mux = 1'b0;             // Do SubBytes
                              
               data_we = 4'hF;
               key_we  = 4'hF;
               
               cnt_up = 1'b1;
               next_state = ENC128_RND;
               
            end // case: ENC128_INIT_RND
          
          
          ENC128_RND: 
            begin
               //Main round
               //Data is read from DATA3 to DATA0
               
               if (decrypt) begin
                  shrow_mux = 2'b01; //Inverse shiftrows
                  sb_mux    = 1'b1;  //No AddroundKey after MixColumns 
               end
               
               cnt_up   = 1'b1;
               
               data_we = 4'hF;
               key_we  = 4'hF;
               
               if(round_cnt == 4'h9) next_state = ENC128_FINAL; 
                
            end // case: ENC128_RND
          
          ENC128_FINAL: 
            begin
               //Final round
               //Addroundkey is performed in this state
               
               addrk_mux = 1'b1;                   // No MixColumns
               sb_mux    = 1'b0;                   // No SubBytes, choose input which causes least toggling in Sbox
               dpout_mux = 1'b1;                   // No SubBytes
               shrow_mux = 2'b10;                  // No ShiftRows

               data_we = 4'hF;
               
               next_state = IDLE;
               done = 1'b1;
            end
          
          default: next_state = IDLE;
    endcase
 end

   
   // *****************************************************************************
   // Sequential part of the FSM
   // *****************************************************************************
   always @ (posedge clk_core, negedge reset_core_n)
     begin
        if (!reset_core_n) 
          begin
             state <= IDLE;
          end
        else if (clken_core) 
          begin
             if (stop)
               state <= IDLE;
             else if (start || running)
               state <= next_state;
          end
     end

   
   // *****************************************************************************
   // Roundcounter
   // *****************************************************************************
   always @ (posedge clk_core, negedge reset_core_n)
     begin
        if (!reset_core_n) 
          begin
             round_cnt <= 'b0;
          end
        else if (clken_core) 
          begin
             if (reset_cnt) 
               round_cnt <= 'b0;
             else if (cnt_up) 
               round_cnt <= round_cnt + 1;
          end
     end

   
   // *****************************************************************************
   // Calculation of round constants
   // *****************************************************************************

   
   always@* 
     begin
        
        case (round_cnt)
          4'h0: 
            begin
               if (decrypt) rcon = 8'h36;
               else         rcon = 8'h01; 
            end

          4'h1: 
            begin
               if (decrypt) rcon = 8'h1b;
               else         rcon = 8'h02; 
            end

          4'h2: 
            begin
               if (decrypt) rcon = 8'h80;
               else         rcon = 8'h04; 
            end

          4'h3: 
            begin
               if (decrypt) rcon = 8'h40;
               else         rcon = 8'h08; 
            end

          4'h4: 
            begin
               if (decrypt) rcon = 8'h20;
               else         rcon = 8'h10; 
            end

          4'h5: 
            begin
               if (decrypt) rcon = 8'h10;
               else         rcon = 8'h20; 
            end

          4'h6: 
            begin
               if (decrypt) rcon = 8'h08;
               else         rcon = 8'h40; 
            end

          4'h7: 
            begin
               if (decrypt) rcon = 8'h04;
               else         rcon = 8'h80; 
            end

          4'h8: 
            begin
               if (decrypt) rcon = 8'h02;
               else         rcon = 8'h1b; 
            end

          4'h9: 
            begin
               if (decrypt) rcon = 8'h01;
               else         rcon = 8'h36; 
            end
            
          default: rcon = 8'h01;
        endcase // case (round_cnt)       
     end
   
endmodule
