// *****************************************************************************
//
// Title       : v1_sbox
// Creator     : Øivind Ekelund
// Date        : 25.03.09
//
// Description : S-box, used in the SubBytes step and Key-expansion
//
// *****************************************************************************

module v1_sbox(
                 input wire [7:0] sb_in, 
                 input wire inverse, 
                 output wire [7:0] sb_out
                 );
   
   // *****************************************************************************
   // Signal declarations
   // *****************************************************************************
   wire [7:0]                      mapped_gf16;
   wire [7:0]                      inv_out;

   
   // *****************************************************************************
   // Logics performing SubBytes
   // *****************************************************************************
   
   //Mux, possible inverse affine and mapping to GF16
   assign mapped_gf16 = (inverse) ? invaffine_map(sb_in) : map_gf16(sb_in);
   
   //Inversion
   assign  inv_out = inversion(mapped_gf16[7:4], mapped_gf16[3:0]);
   
   
   //Mux, inverse mapping and possible affine
   assign sb_out = (inverse) ? invmap(inv_out) : invmap_affine(inv_out);
   
   
   ////////////////////////////////////////////////////////////////////////// 
   // Functions used in the S-box:
   
   //Function performing inverse affine + mapping to GF16
   function automatic [7 : 0] invaffine_map; 
      input [7 : 0]                p; 
      begin 
         invaffine_map[0] = p[2] ^ p[6] ^ p[7];
         invaffine_map[1] = p[1] ^ p[3] ^ p[5]; 
         invaffine_map[2] = p[1] ^ p[2] ^ p[5] ^ p[6] ^ p[7]; 
         invaffine_map[3] = p[5] ^ p[7]; 
         invaffine_map[4] = p[3] ^ p[4] ^ p[5];
         invaffine_map[5] = p[0] ^ p[4] ^ p[5] ^ p[6];
         invaffine_map[6] = p[0] ^ p[1] ^ p[2] ^ p[3] ^ p[6] ^ p[7];
         invaffine_map[7] = p[1] ^ p[2] ^ p[6] ^ p[7];
         
         invaffine_map = invaffine_map ^ 8'h7D;  
      end
   endfunction
   
   //Function performing mapping to GF16
   function automatic [7 : 0] map_gf16; 
      input [7 : 0] p; 
      begin
         map_gf16[0] = p[0] ^ p[1] ^ p[6];
         map_gf16[1] = p[1] ^ p[4] ^ p[6];
         map_gf16[2] = p[1] ^ p[2] ^ p[3] ^ p[4] ^ p[7];
         map_gf16[3] = p[1] ^ p[2] ^ p[6] ^ p[7];
         map_gf16[4] = p[1] ^ p[2] ^ p[3] ^ p[5] ^ p[7];
         map_gf16[5] = p[2] ^ p[3] ^ p[5] ^ p[7];
         map_gf16[6] = p[1] ^ p[2] ^ p[3] ^ p[4] ^ p[6] ^ p[7];
         map_gf16[7] = p[5] ^ p[7];
      end
   endfunction
   
   //Function performing inverse mapping
   function automatic [7 : 0] invmap; 
      input [7 : 0] p; 
      begin
         invmap[0]= p[0] ^ p[2] ^ p[4] ^ p[5] ^ p[6];
         invmap[1]= p[4] ^ p[5];
         invmap[2]= p[1] ^ p[2] ^ p[3] ^ p[4] ^ p[7];
         invmap[3]= p[1] ^ p[2] ^ p[3] ^ p[4] ^ p[5];
         invmap[4]= p[1] ^ p[2] ^ p[4] ^ p[5] ^ p[6];
         invmap[5]= p[1] ^ p[5] ^ p[6];
         invmap[6]= p[2] ^ p[6];
         invmap[7]= p[1] ^ p[5] ^ p[6] ^ p[7];
      end
   endfunction
   
   //Function performing inverse mapping + affine
   function automatic [7 : 0] invmap_affine; 
      input [7 : 0] p; 
      begin
         invmap_affine[0]= p[0] ^ p[1] ^ p[2] ^ p[6] ^ p[7]; 
         invmap_affine[1]= p[0] ^ p[7]; 
         invmap_affine[2]= p[0] ^ p[2] ^ p[3] ^ p[4] ^ p[5] ^ p[6]; 
         invmap_affine[3]= p[0] ^ p[1] ^ p[2]; 
         invmap_affine[4]= p[0] ^ p[1] ^ p[4] ^ p[7];
         invmap_affine[5]= p[2] ^ p[7];
         invmap_affine[6]= p[4] ^ p[5] ^ p[6] ^ p[7];
         invmap_affine[7]= p[2] ^ p[3] ^ p[7];
         
         invmap_affine = invmap_affine ^ 8'h63;
      end
   endfunction
   
   //Function performing multiplication in GF4
   function automatic [1:0] mul_gf4;
      input [1:0 ]p1, p2;
      begin
         mul_gf4[1] = (p1[1] && p2[1]) ^ (p1[1] && p2[0]) ^ (p1[0] && p2[1]); 
         mul_gf4[0] = (p1[1] && p2[1]) ^ (p1[0] && p2[0]);
      end
   endfunction
   
   //Multiplication with lambda in GF16
   function automatic [3:0] mul_l_gf16;
      input [3:0] p;
      begin
         mul_l_gf16[3] = p[2] ^ p[0];  
         mul_l_gf16[2] = p[3] ^ p[2] ^ p[1] ^ p[0];
         mul_l_gf16[1] = p[3];
         mul_l_gf16[0] = p[2];
      end   
   endfunction

   //Squaring in GF16
   function automatic [3:0] square_gf16;
      input [3:0] p;
      begin
         square_gf16[3] = p[3];
         square_gf16[2] = p[3] ^ p[2];
         square_gf16[1] = p[2] ^ p[1];
         square_gf16[0] = p[3] ^ p[1] ^ p[0];
      end   
   endfunction

   //Multiplication with phi in GF4
   function automatic [1:0] mul_f_gf4;
      input [1:0] p;
      begin
         mul_f_gf4[1] = p[1] ^ p[0];
         mul_f_gf4[0] = p[1];
      end   
   endfunction

   //Multiplication in GF16
   function automatic [3:0] mul_gf16;
      input [3:0] p1,p2;

      /*
       // Original function, causes LINT warning
       reg [1:0]   p1l, p1h, p2l, p2h, a, b, c;
       begin
       p1l = p1[1:0];
       p1h = p1[3:2];
       p2l = p2[1:0];
       p2h = p2[3:2];

       a = mul_gf4(p1h, p2h);
       b = mul_gf4((p1h ^ p1l), (p2h ^ p2l));
       c = mul_gf4(p1l, p2l);

       a = mul_f_gf4(a);

       mul_gf16[3:2] =  b ^ c;
       mul_gf16[1:0] =  a ^ c;
      end
       */

      // Equivalent function
      begin
         mul_gf16[3:2] =  mul_gf4((p1[3:2] ^ p1[1:0]), (p2[3:2] ^ p2[1:0])) ^ mul_gf4(p1[1:0], p2[1:0]);
         mul_gf16[1:0] =  mul_f_gf4(mul_gf4(p1[3:2], p2[3:2])) ^ mul_gf4(p1[1:0], p2[1:0]);
      end
   endfunction
   
   //Multiplicative inversion in GF16
   function automatic [3:0] mulinv_gf16;
      input [3:0] x;
      
      begin
         case (x)
           4'h0: mulinv_gf16 = 4'h0;
           4'h1: mulinv_gf16 = 4'h1;
           4'h2: mulinv_gf16 = 4'h3;
           4'h3: mulinv_gf16 = 4'h2;
           4'h4: mulinv_gf16 = 4'hF;
           4'h5: mulinv_gf16 = 4'hC;
           4'h6: mulinv_gf16 = 4'h9;
           4'h7: mulinv_gf16 = 4'hB;
           4'h8: mulinv_gf16 = 4'hA;
           4'h9: mulinv_gf16 = 4'h6;
           4'hA: mulinv_gf16 = 4'h8;
           4'hB: mulinv_gf16 = 4'h7;
           4'hC: mulinv_gf16 = 4'h5;
           4'hD: mulinv_gf16 = 4'hE;
           4'hE: mulinv_gf16 = 4'hD;
           4'hF: mulinv_gf16 = 4'h4;
         endcase
      end
   endfunction

   //Inversion in GF16
   function automatic [7:0] inversion;
      input [3:0] p1, p2;
      
      /*
       //Original function, causes LINT warning
       reg [3:0]   c;
       begin
       c = mul_l_gf16(square_gf16(p1)) ^ mul_gf16(p2, p1 ^ p2);
       c = mulinv_gf16(c);
       inversion[7:4] = mul_gf16(c, p1);
       inversion[3:0] = mul_gf16(c, p1 ^ p2);
      end
       */
      
      //Equivalent function:
      begin
         inversion[7:4] = mul_gf16( mulinv_gf16(mul_l_gf16(square_gf16(p1)) ^ mul_gf16(p2, p1 ^ p2)) , p1);
         inversion[3:0] = mul_gf16( mulinv_gf16(mul_l_gf16(square_gf16(p1)) ^ mul_gf16(p2, p1 ^ p2)) , p1 ^ p2);
      end
      
   endfunction

endmodule

