// *****************************************************************************
//
// Title       : d038_datapath
// Creator     : Øivind Ekelund
// Date        : 25.03.09
//
// Description : Datapath of the AES module
//
// *****************************************************************************

module d038_datapath(
                     input wire [127:0] data_in,
                     input wire [127:0] key_in,
                     input wire sb_inv,
                     input wire mc_inv,
                     input wire addrk_mux,
                     input wire decrypt,
                     input wire [1:0] data_mux,
                     input wire [1:0] key_mux,
                     input wire keymc_mux,
                     input wire sb_mux,
                     input wire [1:0] dpout_mux,
                     input wire [3:0] data_out_mux,
                     input wire [7:0] rcon,
                     output wire [127:0] data_out,
                     output wire [127:0] key_out
                     );
   
   // *****************************************************************************
   // Signal declarations
   // *****************************************************************************
   
   wire [127:0]                          data_invshiftrow;
   wire [127:0]                          data_shiftrow;
   wire [127:0]                          rows_shifted;
   wire [31:0]                           mc_out;
   wire [31:0]                           addrk_out;
   wire [31:0]                           mc_inv_out;
   wire [31:0]                           roundkey;
   wire [31:0]                           key_in_31_0_rot;
   wire [31:0]                           sb_out;
   wire [31:0]                           rcon_out;
   wire [31:0]                           sb_key;
   wire [31:0]                           mc_in;
   wire                                  kexp_mux;
   
   reg [31:0]                            datapath32_out;
   reg [31:0]                            key_mux_out;
   reg [31:0]                            sb_in;
   reg [31:0]                            data_mux_out;

   
   // *****************************************************************************
   // Logics for roundkey computation
   // *****************************************************************************
   
   assign kexp_mux = sb_mux ? decrypt : 1'b1;
   
   d038_keypath 
     u_keypath(
               .key_in(key_in),
               .from_sb(rcon_out),
               .kexp_mux(kexp_mux),
               .key_out(key_out),
               .to_sb(sb_key)
               );
   
   // *****************************************************************************
   // MixColumns and InvMixColumns
   // *****************************************************************************
   
   d038_mixcolumns 
     u_mixcolumns(
                  .mc_in(mc_in),
                  .roundkey(roundkey),
                  .inverse(mc_inv),
                  .mc_out(mc_out)
                  );
   
   // *****************************************************************************
   // InvMixColumns, used on the roundkey during encryption
   // *****************************************************************************
   
   //d038_mixcolumns 
   //  u_invmixcolumns(
   //                  .mc_in(key_mux_out),
   //                  .inverse(1'b1),
   //                  .mc_out(mc_inv_out)
   //                  );

   // *****************************************************************************
   // Sboxes
   // ***************************************************************************** 
   
   d038_sbox 
     u0_sbox(
             .sb_in(sb_in[7:0]),
             .inverse(sb_inv),
             .sb_out(mc_in[7:0])
             );
   d038_sbox 
     u1_sbox(
             .sb_in(sb_in[15:8]),
             .inverse(sb_inv),
             .sb_out(mc_in[15:8])
             );
   d038_sbox 
     u2_sbox(
             .sb_in(sb_in[23:16]),
             .inverse(sb_inv),
             .sb_out(mc_in[23:16])
             );
   d038_sbox 
     u3_sbox(
             .sb_in(sb_in[31:24]),
             .inverse(sb_inv),
             .sb_out(mc_in[31:24])
             );

   // *****************************************************************************
   // Muxes and xor-gates in the datapath
   // ***************************************************************************** 
   
   //Rotate keywords
   assign key_in_31_0_rot = {sb_key[23:16], sb_key[15:8], sb_key[7:0], sb_key[31:24]};
   
   //invShiftrows
   assign data_invshiftrow = 
                             {
                              data_in[127:120], data_in[23:16], data_in[47:40], data_in[71:64],
                              data_in[95:88], data_in[119:112], data_in[15:8], data_in[39:32],
                              data_in[63:56], data_in[87:80], data_in[111:104], data_in[7:0],
                              data_in[31:24], data_in[55:48], data_in[79:72], data_in[103:96]
                              };
   
   //shiftrows
   assign data_shiftrow = 
                          {
                           data_in[127:120], data_in[87:80], data_in[47:40], data_in[7:0],
                           data_in[95:88], data_in[55:48], data_in[15:8], data_in[103:96],
                           data_in[63:56], data_in[23:16], data_in[111:104], data_in[71:64],
                           data_in[31:24], data_in[119:112], data_in[79:72], data_in[39:32]
                           };
   
   //Mux, choosing input to Mixcolumns
   always @ * 
     begin
        case(data_mux)
          2'b00: data_mux_out = data_in[31:0];
          2'b01: data_mux_out = data_in[63:32];
          2'b10: data_mux_out = data_in[95:64];
          2'b11: data_mux_out = data_in[127:96];
        endcase
     end
   
   //Mux, choosing input to the Sbox
   always @ * 
     begin
        case(sb_mux)
          1'b0: sb_in = data_mux_out;
          1'b1: sb_in = key_in_31_0_rot;
        endcase
     end
   
   //Mux, choosing output from the 32-bit datapath
   always @ * 
     begin
        case(dpout_mux)
          2'b00: datapath32_out = sb_in;
          2'b01: datapath32_out = mc_in;
          2'b10: datapath32_out = mc_out;
          default: datapath32_out = sb_in;
        endcase
     end
   
   //Mux, choosing roundkey
   always @ * 
     begin
        case(key_mux)
          2'b00: key_mux_out = key_in[31:0];
          2'b01: key_mux_out = key_in[63:32];
          2'b10: key_mux_out = key_in[95:64];
          2'b11: key_mux_out = key_in[127:96];
        endcase
     end
   
   //Mux, choosing roundkey
   assign roundkey = key_mux_out;
  
   //Mux, choosing between shiftrows and invshiftrows
   assign rows_shifted = (decrypt) ? data_invshiftrow : data_shiftrow;
   
   //Muxes, chossing (inv)shiftrow or output form datapath32
   assign data_out[31:0] = (data_out_mux[0]) ? addrk_out : rows_shifted[31:0];
   assign data_out[63:32] = (data_out_mux[1]) ? addrk_out : rows_shifted[63:32];
   assign data_out[95:64] = (data_out_mux[2]) ? addrk_out : rows_shifted[95:64];
   assign data_out[127:96] = (data_out_mux[3]) ? addrk_out : rows_shifted[127:96];
   
   //Addroundkey, if InvMixColumns is NOT applied
   assign addrk_out = (decrypt && (dpout_mux==2'b10)) ? datapath32_out : (datapath32_out ^ roundkey);
   
   //Add roundconstand
   assign rcon_out = mc_in ^ {rcon, 24'h000000};
   
endmodule
