// *****************************************************************************
//
// Title       : v5_aes_datapath
// Creator     : Øivind Ekelund
// Date        : 25.03.09
//
// Description : Datapath of the AES module
//               Datapath is split into two main parts data and key.
//               The data part operates on 32 bits at a time, and performs the
//               4 AES operations, ShiftRows, SubBytes (sboxes), MixColumns and
//               AddRoundKey(xor)
//               The key path also utilizes the sboxes in the rounds where it is
//               not used by the data path. The round constant is changed each
//               round and used in the key expansion.
//
// *****************************************************************************

module v5_aes_datapath
  (
   input wire [127:0]     data_in,
   input wire [255:0]     key_in,
   input wire             sb_inv,
   input wire             mc_inv,
   input wire             addrk_mux,
   input wire             decrypt,
   input wire [1:0]       data_mux,
   input wire [2:0]       key_mux,
   input wire [1:0]       sb_mux,
   input wire [1:0]       dpout_mux,
   input wire [3:0]       data_out_mux,
   input wire [7:0]       rcon,
   output wire [127:0]    data_out,
   output wire [255:0]    key_out
   );
   
   // *****************************************************************************
   // Signal declarations
   // *****************************************************************************
   
   wire [127:0]           data_invshiftrow;
   wire [127:0]           data_shiftrow;
   wire [127:0]           rows_shifted;
   wire [31:0]            mc_out;
   wire [31:0]            addrk_in;
   wire [31:0]            addrk_out;
   wire [31:0]            mc_inv_out;
   wire [31:0]            key_in_31_0_rot;
   wire [31:0]            sb_out;
   wire [31:0]            rcon_out;
   
   reg [31:0]             datapath32_out;
   reg [31:0]             mc_in;
   reg [31:0]             key_mux_out;
   reg [31:0]             sb_in;
   
   
   // *****************************************************************************
   // Logics for roundkey computation
   // *****************************************************************************
   
   v5_aes_keypath 
     u_keypath(
               .key_in   (key_in),
               .from_sb  (datapath32_out),
               .key_out  (key_out)
               );
   
   // *****************************************************************************
   // MixColumns and InvMixColumns
   // *****************************************************************************
   
   v5_aes_mixcolumns 
     u_mixcolumns(
                  .mc_in   (mc_in),
                  .roundkey(key_mux_out),
                  .inverse (mc_inv),
                  .mc_out  (mc_out)
                  );

   // *****************************************************************************
   // Sboxes
   // ***************************************************************************** 
   
   v5_aes_sbox 
     u0_sbox(
             .sb_in   (sb_in[7:0]),
             .inverse (sb_inv),
             .sb_out  (sb_out[7:0])
             );
   v5_aes_sbox 
     u1_sbox(
             .sb_in   (sb_in[15:8]),
             .inverse (sb_inv),
             .sb_out  (sb_out[15:8])
             );
   v5_aes_sbox 
     u2_sbox(
             .sb_in(sb_in[23:16]),
             .inverse(sb_inv),
             .sb_out(sb_out[23:16])
             );
   v5_aes_sbox 
     u3_sbox(
             .sb_in   (sb_in[31:24]),
             .inverse (sb_inv),
             .sb_out  (sb_out[31:24])
             );

   // *****************************************************************************
   // Muxes and xor-gates in the datapath
   // ***************************************************************************** 
   
   //Rotate keywords
   assign key_in_31_0_rot = {key_in[23:16], key_in[15:8], key_in[7:0], key_in[31:24]};
   
   //invShiftrows
   assign data_invshiftrow = 
                             {
                              data_in[127:120], data_in[23:16],   data_in[47:40],   data_in[71:64],
                              data_in[95:88],   data_in[119:112], data_in[15:8],    data_in[39:32],
                              data_in[63:56],   data_in[87:80],   data_in[111:104], data_in[7:0],
                              data_in[31:24],   data_in[55:48],   data_in[79:72],   data_in[103:96]
                              };
   
   //shiftrows
   assign data_shiftrow = 
                          {
                           data_in[127:120], data_in[87:80],   data_in[47:40],   data_in[7:0],
                           data_in[95:88],   data_in[55:48],   data_in[15:8],    data_in[103:96],
                           data_in[63:56],   data_in[23:16],   data_in[111:104], data_in[71:64],
                           data_in[31:24],   data_in[119:112], data_in[79:72],   data_in[39:32]
                           };
   
   //Mux, choosing input to Mixcolumns
   always @ * 
     begin
        case(data_mux)
          2'b00: mc_in = data_in[31:0];
          2'b01: mc_in = data_in[63:32];
          2'b10: mc_in = data_in[95:64];
          2'b11: mc_in = data_in[127:96];
        endcase
     end
   
   //Mux, choosing input to the Sbox
   always @ * 
     begin
        case(sb_mux)
          2'b00: sb_in   = addrk_out;
          2'b01: sb_in   = addrk_in;
          2'b10: sb_in   = key_in_31_0_rot;
          2'b11: sb_in   = key_in[159:128];
        endcase
     end
   
   //Mux, choosing output from the 32-bit datapath
   always @ * 
     begin
        case(dpout_mux)
          2'b00: datapath32_out   = sb_out;
          2'b01: datapath32_out   = addrk_out;
          2'b10: datapath32_out   = rcon_out;
          default: datapath32_out = sb_out;
        endcase
     end
   
   //Mux, choosing key word for roundkey
   always @ * 
     begin
        case(key_mux)
          3'b000: key_mux_out = key_in[31:0];
          3'b001: key_mux_out = key_in[63:32];
          3'b010: key_mux_out = key_in[95:64];
          3'b011: key_mux_out = key_in[127:96];
          3'b100: key_mux_out = key_in[159:128];
          3'b101: key_mux_out = key_in[191:160];
          3'b110: key_mux_out = key_in[223:192];
          3'b111: key_mux_out = key_in[255:224];
        endcase
     end
   
   //Mux, choosing input to Addroundkey
   assign addrk_in = addrk_mux ? mc_in : mc_out;
   
   //Mux, choosing between shiftrows and invshiftrows
   assign rows_shifted = decrypt ? data_invshiftrow : data_shiftrow;
   
   //Muxes, choosing (inv)shiftrow or output from datapath32
   assign data_out[31:0]   = data_out_mux[0] ? datapath32_out : rows_shifted[31:0];
   assign data_out[63:32]  = data_out_mux[1] ? datapath32_out : rows_shifted[63:32];
   assign data_out[95:64]  = data_out_mux[2] ? datapath32_out : rows_shifted[95:64];
   assign data_out[127:96] = data_out_mux[3] ? datapath32_out : rows_shifted[127:96];
   
   //Addroundkey
   assign addrk_out = addrk_in ^ key_mux_out;
   
   //Add roundconstand
   assign rcon_out = sb_out ^ {rcon, 24'h000000};
   
endmodule
