// *****************************************************************************
//
// Title       : v5_aes_sequencer
// Creator     : Øivind Ekelund
// Date        : 25.03.09
//
// Description : Sequencer, controlling the AES datapath and registers
//
// *****************************************************************************

module v5_aes_sequencer
  (
   input wire          clk_core,
   input wire          clken_core,
   input wire          reset_core_n,
   input wire          start,
   input wire          decrypt,
   input wire          aes256,
   input wire          stop,
   input wire          keybufen,
   output reg          sb_inv,
   output reg          mc_inv,
   output reg          running,
   output reg          addrk_mux,
   output reg [1:0]    data_mux,
   output reg [2:0]    key_mux,
   output reg [1:0]    sb_mux,
   output reg [1:0]    dpout_mux,
   output reg [3:0]    data_out_mux,
   output reg [7:0]    rcon,
   output reg [3:0]    data_we,
   output reg [7:0]    key_we,
   output reg          done
   );
   
   // *****************************************************************************
   // States in the FSM
   // *****************************************************************************
   
   parameter [4:0]

                IDLE             = 5'b00000, 
                 
                ENC128_INIT_RND  = 5'b00010,
                ENC128_SHROW     = 5'b00011,
                ENC128_RND       = 5'b00100,
                ENC128_FINAL     = 5'b00101,

                DEC128_INIT_RND  = 5'b00001,   
                DEC128_SHROW     = 5'b00110,
                DEC128_RND       = 5'b00111,
                DEC128_FINAL     = 5'b01000,

                ENC256_INIT_RND  = 5'b10000,
                ENC256_SHROW1    = 5'b10001,
                ENC256_RND1      = 5'b10010,
                ENC256_SHROW2    = 5'b10011,
                ENC256_FINAL     = 5'b10100,
                ENC256_RND2      = 5'b10101,

                DEC256_INIT_RND  = 5'b11000,
                DEC256_SHROW1    = 5'b11001,
                DEC256_RND1      = 5'b11010,
                DEC256_SHROW2    = 5'b11011,
                DEC256_RND2      = 5'b11100,
                DEC256_FINAL     = 5'b11101;

   // *****************************************************************************
   // Register- and signal declarations
   // *****************************************************************************
   
   reg [4:0]           state;
   reg [4:0]           next_state;
   reg [5:0]           round_cnt;
   reg                 cnt_up;
   reg                 reset_cnt; 
   
   // *****************************************************************************
   // Combinational part of the FSM
   // *****************************************************************************
   always @ * 
     begin
        //default values:
        //Controlsignals for the datapath
        sb_inv       = 1'b0;
        mc_inv       = 1'b0;
        running      = 1'b1;
        addrk_mux    = 1'b0;
        data_mux     = 2'b00;
        key_mux      = 3'b000;
        sb_mux       = 2'b00;
        dpout_mux    = 2'b00;
        data_out_mux = 4'h0;     
        data_we      = 4'h0;
        key_we       = 8'h00;
        done         = 1'b0;
        
        //Internal controlsignals
        next_state = state;
        cnt_up     = 1'b0;
        reset_cnt  = 1'b0;
        
        case (state)
          IDLE: 
            begin
               sb_mux    = 2'b11;  // Input to Sbox: KEY4, to minimize toggling in Sbox if dataregister is read/written
               running   = 1'b0;
               
               if (start)
                 begin
                    reset_cnt = 1'b1; 
                    
                    if (!aes256 && !decrypt)
                      next_state = ENC128_INIT_RND;
                    else if (!aes256 && decrypt)
                      next_state = DEC128_INIT_RND;
                    else if (aes256 && !decrypt)
                      next_state = ENC256_INIT_RND;
                    else if (aes256 && decrypt)
                      next_state = DEC256_INIT_RND;
                 end
            end
          
          ENC128_INIT_RND: 
            begin
               //Initial round of AES128 encryption
               //Data is read from DATA3 to DATA0
               //Key is read from KEY7 to KEY4 if keybuffering is enabled,
               //if else, key is read from KEY3 to KEY0
               //SubBytes and Addroundkey is perfomed in this state
               
               sb_mux    = 2'b00;            // Do Addroundkey
               addrk_mux = 1'b1;             // No MixColumns
               dpout_mux = 2'b00;            // Do SubBytes
               data_mux  = ~round_cnt[1:0];
               cnt_up    = 1'b1;
               
               if (keybufen)
                 //Read key from KEYH
                 key_mux = ~round_cnt[2:0];
               else
                 //Read key from KEYL
                 key_mux = {1'b0,~round_cnt[1:0]};
               
               if (round_cnt[1:0] == 2'b11)
                 begin
                    data_out_mux = 4'b0001; 
                    data_we      = 4'b0001;
                    next_state   = ENC128_SHROW;
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'b0010;
                    data_we      = 4'b0010;
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'b0100;
                    data_we      = 4'b0100;
                 end 
               else if (round_cnt[1:0] == 2'b00) 
                 begin 
                    data_out_mux = 4'b1000; 
                    data_we      = 4'b1000;
                 end
            end // case: ENC128_INIT_RND
          
          ENC128_SHROW: 
            begin
               //Shiftrows is performed while the sbox is used in keygeneration
               //KEY3 is expanded in this state
               
               data_we   = 4'b1111;      // Write to entire dataregister
               key_we    = 8'b00001000;  // Write to KEY3
               sb_mux    = 2'b10;        // Input to Sbox: KEY0 rotated
               dpout_mux = 2'b10;        // Add roundconstant
               addrk_mux = 1'b1;         // No MixColumns, minimize toggling in Addroundkey
               
               if (round_cnt == 6'h28)
                 next_state = ENC128_FINAL;
               else
                 next_state = ENC128_RND;
            end
          
          ENC128_RND: 
            begin
               //Main round
               //Data is read from DATA3 to DATA0
               //Key is read from KEY3 to KEY0
               //Addroundkey, MixColumns and SubBytes is performed in this state
               //KEY2 - KEY0 is expanded in this state
               
               data_mux = ~round_cnt[1:0];
               key_mux  = {1'b0,~round_cnt[1:0]};
               cnt_up   = 1'b1;
               
               if(round_cnt[1:0] == 2'b11)
                 begin
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;
                    next_state   = ENC128_SHROW;
                 end 
               else if(round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;
                    key_we       = 8'h01;
                 end 
               else if(round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;
                    key_we       = 8'h02;
                 end 
               else if(round_cnt[1:0] == 2'b00) 
                 begin 
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                    key_we       = 8'h04;
                 end
            end // case: ENC128_RND
          
          ENC128_FINAL: 
            begin
               //Final round
               //Data is read from DATA3 to DATA0
               //Key is read from KEY3 to KEY0
               //Addroundkey is performed in this state
               //KEY2 - KEY0 is expanded in this state
               
               addrk_mux = 1'b1;                    // No MixColumns
               sb_mux    = 2'b11;                   // No SubBytes, choose input which causes least toggling in Sbox
               dpout_mux = 2'b01;                   // Do Addroundkey
               data_mux  = ~round_cnt[1:0];
               key_mux   = {1'b0,~round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if(round_cnt[1:0] == 2'b11)
                 begin
                    data_out_mux = 4'b0001; 
                    data_we      = 4'b0001;
                    done         = 1'b1;
                    next_state   = IDLE;
                 end 
               else if(round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'b0010; 
                    data_we      = 4'b0010;
                    key_we       = 8'h01;
                 end
               else if(round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'b0100;
                    data_we      = 4'b0100;
                    key_we       = 8'h02;
                 end 
               else if(round_cnt[1:0] == 2'b00) 
                 begin 
                    data_out_mux = 4'b1000; 
                    data_we      = 4'b1000;
                    key_we       = 8'h04;
                 end
            end
          
          DEC128_INIT_RND: 
            begin
               //Initial round of AES128 decryption
               //Data is read from DATA0 to DATA3
               //Key is read from KEY4 to KEY7 if keybuffering is enabled,
               //if else, key is read from KEY0 to KEY3
               //Addroundkey and inverse SubBytes is performed in this round
               //KEY0 - KEY2 is expanded in this state
               
               sb_inv    = 1'b1;            // Do inverse SubBytes
               sb_mux    = 2'b00;           // Do Addroundkey
               addrk_mux = 1'b1;            // No MixColumns
               dpout_mux = 2'b00;           // Choose output from SubBytes
               data_mux  = round_cnt[1:0];
               cnt_up    = 1'b1;
               
               if (keybufen)
                 //Read key from KEYH
                 key_mux = {1'b1,round_cnt[1:0]};
               else
                 //Read key from KEYL
                 key_mux = {1'b0,round_cnt[1:0]};
               
               if(round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;
                    key_we       = 8'h01;
                 end 
               else if(round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;
                    key_we       = 8'h02;
                 end
               else if(round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;
                    key_we       = 8'h04;
                 end 
               else if(round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h8; 
                    data_we      = 4'h8;
                    next_state   = DEC128_SHROW;
                 end
            end
          
          DEC128_SHROW: 
            begin
               //Invshiftrows is performed while the sbox is used in keygeneration
               //KEY3 is expanded in this state
               
               data_we   = 4'hF;     // Write to entire dataregister
               key_we    = 8'h08;    // Write to KEY3
               sb_mux    = 2'b10;    // Input to Sbox: KEY0 rotated
               dpout_mux = 2'b10;    // Add roundconstant
               addrk_mux = 1'b1;     // No MixColumns, minimize toggling in Addroundkey
               
               if (round_cnt == 6'h28)
                 next_state = DEC128_FINAL;
               else
                 next_state = DEC128_RND;
            end
          
          DEC128_RND: 
            begin
               //Main round
               //Data is read from DATA0 to DATA3
               //Key is read from KEY0 to KEY3 
               //Addroundkey, inverse MixColumns and inverse SubBytes is performed in this state
               //KEY0 - KEY2 is expanded in this state
               
               sb_inv    = 1'b1;                  // Do inverse SubBytes
               sb_mux    = 2'b01;                 // No addroundkey after MixColumns
               mc_inv    = 1'b1;                  // Do inverse MixColumns
               data_mux  = round_cnt[1:0];
               key_mux   = {1'b0,round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'b0001;
                    data_we      = 4'b0001;
                    key_we       = 8'h01;
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'b0010;
                    data_we      = 4'b0010;
                    key_we       = 8'h02;
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'b0100;
                    data_we      = 4'b0100;
                    key_we       = 8'h04;
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'b1000; 
                    data_we      = 4'b1000;
                    next_state   = DEC128_SHROW;
                 end
            end
          
          DEC128_FINAL: 
            begin
               //Final round
               //Data is read from DATA0 to DATA3
               //Key is read from KEY0 to KEY3
               //Addroundkey is performed in this state 
               
               addrk_mux = 1'b1;                  // No MixColumns
               sb_mux    = 2'b11;                 // No SubBytes, choose input which causes least toggling in Sbox
               dpout_mux = 2'b01;                 // Do Addroundkey
               data_mux  = round_cnt[1:0];
               key_mux   = {1'b0,round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'b0001;
                    data_we      = 4'b0001;
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'b0010;
                    data_we      = 4'b0010; 
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'b0100;
                    data_we      = 4'b0100; 
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'b1000;
                    data_we      = 4'b1000;
                    done         = 1'b1;
                    next_state   = IDLE;
                 end
            end
          
          ENC256_INIT_RND: 
            begin
               //Initial round of AES256 encryption
               //Data is read from DATA3 to DATA0
               //Key is read from KEY7 to KEY4
               //Addroundkey and SubBytes is performed in this state 
               
               addrk_mux = 1'b1;             // No MixColumns
               data_mux  = ~round_cnt[1:0];
               key_mux   = ~round_cnt[2:0];
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                 end 
               else if (round_cnt[1:0] == 2'b01)
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4; 
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2; 
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;
                    next_state   = ENC256_SHROW1;
                 end
            end
          
          ENC256_SHROW1: 
            begin
               //Shiftrows is performed while the sbox is used in keygeneration
               //KEY7 is expanded in this state
               
               data_we    = 4'hF;        // Write to entire dataregister
               key_we     = 8'h80;       // Write to KEY7
               sb_mux     = 2'b10;       // Input to Sbox: KEY0 rotated
               dpout_mux  = 2'b10;       // Add roundconstant
               addrk_mux = 1'b1;         // No MixColumns, minimize toggling in Addroundkey
               
               next_state = ENC256_RND1; 
            end
          
          ENC256_RND1: 
            begin
               //Part of the main round
               //Data is read from DATA3 to DATA0
               //Key is read from KEY3 to KEY0 
               //MixColumns, Addroundkey and SubBytes is performed in this state
               //KEY6 - KEY4 is expanded in this state
               
               data_mux = ~round_cnt[1:0];
               key_mux  = {1'b0, ~round_cnt[1:0]};
               cnt_up   = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                    key_we       = 8'h40;  
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;
                    key_we       = 8'h20;  
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;
                    key_we       = 8'h10;  
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h1; 
                    data_we      = 4'h1;  
                    next_state   = ENC256_SHROW2; 
                 end 
            end
          
          ENC256_SHROW2: 
            begin
               //Shiftrows is performed while the sbox is used in keygeneration
               //KEY3 is expanded in this state
               
               data_we   = 4'hF;    // Write to entire dataregister
               key_we    = 8'h08;   // Write to KEY3
               sb_mux    = 2'b11;   // Input to Sbox: KEY4
               dpout_mux = 2'b00;   // Do not add roundconstant
               addrk_mux = 1'b1;    // No MixColumns, minimize toggling in Addroundkey
               
               if (round_cnt == 6'h38) 
                 begin
                    next_state = ENC256_FINAL;
                    key_we     = 8'h00;
                 end  
               else
                 next_state = ENC256_RND2;    
            end
          
          ENC256_RND2: 
            begin
               //Part of the main round
               //Data is read from DATA3 to DATA0
               //Key is read from KEY7 to KEY4 
               //MixColumns, Addroundkey and SubBytes is performed in this state
               //KEY2 - KEY0 is expanded in this state
               
               data_mux = ~round_cnt[1:0];
               key_mux  = ~round_cnt[2:0];
               cnt_up   = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                    key_we       = 8'h04;  
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;
                    key_we       = 8'h02;  
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;
                    key_we       = 8'h01;  
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;
                    next_state   = ENC256_SHROW1;
                 end    
            end
          
          ENC256_FINAL: 
            begin
               //Final round
               //Data is read from DATA3 to DATA0
               //Key is read from KEY7 to KEY4
               //Addroundkey is performed in this state
                
               addrk_mux = 1'b1;            // No MixColumns
               sb_mux    = 2'b11;           // No SubBytes, choose input which causes least toggling in Sbox
               dpout_mux = 2'b01;           // Do Addroundkey
               cnt_up    = 1'b1;
               data_mux  = ~round_cnt[1:0];
               key_mux   = ~round_cnt[2:0];
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;
                    done         = 1'b1;
                    next_state   = IDLE;
                 end
            end

          DEC256_INIT_RND: 
            begin
               //Initial round of AES256 decryption
               //Data is read from DATA0 to DATA3
               //Key is read from KEY4 to KEY7
               //Addroundkey and Inverse SubBytes is perfomed in this state
               
               sb_inv    = 1'b1;                   // Do inverse SubBytes
               addrk_mux = 1'b1;                   // No MixColumns
               data_mux  = round_cnt[1:0];
               key_mux   = {1'b1, round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h1;
                    data_we      = 4'h1; 
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                    next_state   = DEC256_SHROW1;
                 end
            end // case: DEC256_INIT_RND

          DEC256_SHROW1: 
            begin
               //Invshiftrows is performed while the sbox is used in keygeneration
               //KEY4 - KEY7 is expanded in this state
               
               data_we    = 4'b1111;      // Write to entire dataregister
               sb_mux     = 2'b10;        // Input to Sbox: KEY0 rotated
               dpout_mux  = 2'b10;        // Add roundconstant
               key_we     = 8'hF0;        // Write to KEY4 - KEY7
               addrk_mux = 1'b1;          // No MixColumns, minimize toggling in Addroundkey
               
               next_state = DEC256_RND1;
               
            end
          
          DEC256_RND1: 
            begin
               //Part of the main round
               //Data is read from DATA0 to DATA3
               //Key is read from KEY0 to KEY3
               //Addroundkey, inverse MixColumns and inverse SubBytes is performed in this state
               
               sb_mux    = 2'b01;                   // No Addroundkey after MixColumns
               sb_inv    = 1'b1;                    // Do inverse SubBytes
               mc_inv    = 1'b1;                    // Do inverse MixColumns
               data_mux  = round_cnt[1:0];
               key_mux   = {1'b0, round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;  
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;  
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;  
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h8;
                    data_we      = 4'h8; 
                    next_state   = DEC256_SHROW2;
                 end
            end

          DEC256_SHROW2: 
            begin
               //Invshiftrows is performed while the sbox is used in keygeneration
               //KEY0 - KEY3 is expanded in this state
               
               data_we = 4'hF;    // Write to entire dataregister
               key_we  = 8'h0F;   // Write to KEY0 - KEY3
               sb_mux  = 2'b11;   // Input to Sbox: KEY4
               addrk_mux = 1'b1;  // No MixColumns, minimize toggling in Addroundkey
               
               if (round_cnt == 6'h38) 
                 begin
                    next_state = DEC256_FINAL;
                    key_we     = 8'h00; 
                 end  
               else
                 next_state = DEC256_RND2;   
            end

          DEC256_RND2: 
            begin
               //Part of the main round
               //Data is read from DATA0 to DATA3
               //Key is read from KEY4 to KEY7
               //Addroundkey, inverse MixColumns and inverse SubBytes is performed in this state
               
               sb_mux    = 2'b01;                   // No Addroundkey after MixColumns
               sb_inv    = 1'b1;                    // Do inverse SubBytes
               mc_inv    = 1'b1;                    // Do inverse MixColumns
               data_mux  = round_cnt[1:0];
               key_mux   = {1'b1, round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;  
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2;  
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4;  
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                    next_state   = DEC256_SHROW1;
                 end
            end

          DEC256_FINAL: 
            begin
               //Final round
               //Data is read from DATA0 to DATA3
               //Key is read from KEY4 to KEY7
               //Addroundkey is performed in this state
               
               addrk_mux = 1'b1;                  // No MixColumns
               sb_mux    = 2'b11;                 // No SubBytes, choose input which causes least toggling in Sbox
               dpout_mux = 2'b01;                 // Do Addroundkey
               data_mux  = round_cnt[1:0];
               key_mux   = {1'b1,round_cnt[1:0]};
               cnt_up    = 1'b1;
               
               if (round_cnt[1:0] == 2'b00)
                 begin
                    data_out_mux = 4'h1;
                    data_we      = 4'h1;
                 end 
               else if (round_cnt[1:0] == 2'b01) 
                 begin 
                    data_out_mux = 4'h2;
                    data_we      = 4'h2; 
                 end 
               else if (round_cnt[1:0] == 2'b10) 
                 begin 
                    data_out_mux = 4'h4;
                    data_we      = 4'h4; 
                 end 
               else if (round_cnt[1:0] == 2'b11) 
                 begin 
                    data_out_mux = 4'h8;
                    data_we      = 4'h8;
                    done         = 1'b1;
                    next_state   = IDLE;
                 end
            end
          
          default: 
            next_state = IDLE;
          
        endcase
     end

   
   // *****************************************************************************
   // Sequential part of the FSM
   // *****************************************************************************
   always @ (posedge clk_core, negedge reset_core_n)
     begin
        if (!reset_core_n) 
          begin
             state <= IDLE;
          end
        else if (clken_core) 
          begin
             if (stop)
               state <= IDLE;
             else if (start || running)
               state <= next_state;
          end
     end

   
   // *****************************************************************************
   // Roundcounter
   // *****************************************************************************
   always @ (posedge clk_core, negedge reset_core_n)
     begin
        if (!reset_core_n) 
          begin
             round_cnt <= 'b0;
          end
        else if (clken_core) 
          begin
             if (reset_cnt) 
               round_cnt <= 'b0;
             else if (cnt_up) 
               round_cnt <= round_cnt + 1;
          end
     end

   
   // *****************************************************************************
   // Calculation of round constants
   // *****************************************************************************

   
   always@* 
     begin
        
        case (round_cnt[5:2])
          4'h1: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h01;
               else if (!aes256 && decrypt) 
                 rcon = 8'h36;
               else if (aes256 && !decrypt) 
                 rcon = 8'h01;
               else 
                 rcon = 8'h40;   
            end
          4'h2: 
            begin 
               if (!aes256 && !decrypt) 
                 rcon = 8'h02;
               else if (!aes256 && decrypt) 
                 rcon = 8'h1b;
               else if (aes256 && !decrypt) 
                 rcon = 8'h02;
               else 
                 rcon = 8'h20; 
            end
          4'h3: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h04;
               else if (!aes256 && decrypt) 
                 rcon = 8'h80;
               else if (aes256 && !decrypt) 
                 rcon = 8'h02;
               else 
                 rcon = 8'h20; 
            end
          4'h4: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h08;
               else if (!aes256 && decrypt) 
                 rcon = 8'h40;
               else if (aes256 && !decrypt) 
                 rcon = 8'h04;
               else 
                 rcon = 8'h10; 
            end
          4'h5: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h10;
               else if (!aes256 && decrypt) 
                 rcon = 8'h20;
               else if (aes256 && !decrypt) 
                 rcon = 8'h04;
               else 
                 rcon = 8'h10; 
            end
          4'h6: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h20;
               else if (!aes256 && decrypt) 
                 rcon = 8'h10;
               else if (aes256 && !decrypt) 
                 rcon = 8'h08;
               else 
                 rcon = 8'h08; 
            end
          4'h7: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h40;
               else if (!aes256 && decrypt) 
                 rcon = 8'h08;
               else if (aes256 && !decrypt) 
                 rcon = 8'h08;
               else 
                 rcon = 8'h08; 
            end
          4'h8: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h80;
               else if (!aes256 && decrypt) 
                 rcon = 8'h04;
               else if (aes256 && !decrypt) 
                 rcon = 8'h10;
               else 
                 rcon = 8'h04; 
            end
          4'h9: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h1b;
               else if (!aes256 && decrypt) 
                 rcon = 8'h02;
               else if (aes256 && !decrypt) 
                 rcon = 8'h10;
               else 
                 rcon = 8'h04; 
            end
          4'hA: 
            begin
               if (!aes256 && !decrypt) 
                 rcon = 8'h36;
               else if (!aes256 && decrypt) 
                 rcon = 8'h01;
               else if (aes256 && !decrypt) 
                 rcon = 8'h20;
               else 
                 rcon = 8'h02; 
            end
          ////
          4'hB: 
            begin
               if (decrypt) 
                 rcon = 8'h02;
               else 
                 rcon = 8'h20;
            end
          4'hC: 
            begin
               if (decrypt) 
                 rcon = 8'h01;
               else 
                 rcon = 8'h40;
            end
          4'hD: 
            begin
               if (decrypt) 
                 rcon = 8'h01;
               else 
                 rcon = 8'h40;
            end

          default: rcon = 8'h01;
        endcase // case (round_cnt)       
     end
   
endmodule
